def pga_signature(b):
    """0(...+)"""
    return 0 if b == 0 else 1

def sta_signature(b):
    """+(...-)"""
    return 1 if b == 0 else -1

def stap_signature(b):
    """0+(...-)"""
    return 0 if b == 0 else (1 if b == 1 else -1)

def cga_signature(b):
    """+-(...+)"""
    return 1 if b == 0 else (-1 if b == 1 else 1)

def positive_signature(b):
    """(...+)"""
    return 1

def negative_signature(b):
    """(...-)"""
    return -1

def null_signature(b):
    """(...0)"""
    return 0


import jax
import jax.numpy as jnp


def get_mv_keep_nonzero(a_blade_indices, a_values):
    out_blade_indices = []
    out_a_indices = []
    for i, a_index in enumerate(a_blade_indices):
        if not jnp.allclose(a_values[i], 0):
            out_blade_indices.append(a_index)
            out_a_indices.append(i)
    out_a_indices = jnp.array(out_a_indices, dtype=jnp.int32)

    def _values_mv_keep_nonzero(a_values):
        return a_values[out_a_indices]

    _values_mv_keep_nonzero_jit = jax.jit(_values_mv_keep_nonzero)

    return _values_mv_keep_nonzero_jit, tuple(out_blade_indices)



from ..ops.multiply import get_mv_multiply
from ..signatures import positive_signature
from typing import Sequence, Callable
import flax.linen as nn
import jax.numpy as jnp


class MVProdDense(nn.Module):
    x_ind: Sequence[Sequence[int]]
    w_ind: Sequence[Sequence[int]]
    units: int
    w_init: Callable = nn.initializers.lecun_normal()
    signature: Callable[[int], float] = positive_signature
    prod: str = "gp"

    def setup(self):
        self.mv_multiply, self.result_indices = get_mv_multiply(
            self.x_ind, self.w_ind, signature=self.signature, prod=self.prod
        )

    @nn.compact
    def __call__(self, x):
        w = self.param(
            "w", self.w_init,
            (len(self.w_ind), x.shape[-1], self.units)
        )
        x = jnp.expand_dims(x, -1)
        # [M, ...B, I] * [M, I, O] -> [M, ...B, I, O] -> [M, ...B, O]
        result_values_unreduced = self.mv_multiply(x, w)
        result = jnp.sum(result_values_unreduced, axis=-2)
        return result



import jax
import jax.numpy as jnp
from .reduce_same import get_mv_reduce_same
from functools import cache


@cache
def get_mv_add(a_blade_indices, b_blade_indices):
    out_blade_indices = a_blade_indices + b_blade_indices
    mv_reduce_same, out_blade_indices = get_mv_reduce_same(out_blade_indices)

    def _values_mv_reduce_same(a_values, b_values):
        return mv_reduce_same(jnp.concatenate([a_values, b_values], axis=0))
    _values_mv_add_jit = jax.jit(_values_mv_reduce_same)

    return _values_mv_add_jit, out_blade_indices



import jax
import jax.numpy as jnp
from ..jaxga import dual_blade_index
from functools import cache


@cache
def get_mv_dual(a_blade_indices, dims):
    out_blade_indices = []
    signs = jnp.empty([len(a_blade_indices)], dtype=jnp.float32)
    for i, blade_index in enumerate(a_blade_indices):
        sign, dual_index = dual_blade_index(blade_index, dims)
        out_blade_indices.append(dual_index)
        signs = signs.at[i].set(sign)

    def _values_mv_dual(a_values):
        return a_values * signs

    _values_mv_dual_jit = jax.jit(_values_mv_dual)

    return _values_mv_dual_jit, tuple(out_blade_indices)



import jax
import jax.numpy as jnp
from .multiply import get_mv_multiply
from .add import get_mv_add
from .select import get_mv_select
from ..jaxga import is_scalar_index
from functools import cache


@cache
def get_mv_simple_exp(a_blade_indices, signature):
    mv_multiply, a_sq_indices = get_mv_multiply(
        a_blade_indices, a_blade_indices, signature
    )
    mv_select_scalar, scalar_indices = get_mv_select(
        a_sq_indices, is_scalar_index
    )
    mv_add, out_indices = get_mv_add(scalar_indices, a_blade_indices)

    def _values_mv_simple_exp(a_values):
        a_sq_values = mv_select_scalar(mv_multiply(a_values, a_values))
        a_sq_sqrt = jnp.sign(a_sq_values) * jnp.sqrt(jnp.abs(a_sq_values))

        out_scalar = jnp.where(
            a_sq_sqrt < 0,
            jnp.cos(a_sq_sqrt),
            jnp.where(
                a_sq_sqrt > 0,
                jnp.cosh(a_sq_sqrt),
                jnp.ones_like(a_sq_sqrt)
            )
        )

        out_blade = jnp.where(
            a_sq_sqrt < 0,
            a_values / a_sq_sqrt * jnp.sin(a_sq_sqrt),
            jnp.where(
                a_sq_sqrt > 0,
                a_values / a_sq_sqrt * jnp.sinh(a_sq_sqrt),
                a_values
            )
        )

        return mv_add(out_scalar, out_blade)

    _values_mv_simple_exp_jit = jax.jit(_values_mv_simple_exp)

    return _values_mv_simple_exp_jit, out_indices



import jax
import jax.numpy as jnp

from jaxga.signatures import positive_signature
from ..jaxga import reduce_bases
from functools import cache


@cache
def get_mv_reduce_same(a_blade_indices):
    blade_to_index = {}
    blade_to_blade_index = {}

    indices = []
    unique_count = len(set(a_blade_indices))
    out_indices = [[] for _ in range(unique_count)]
    out_signs = [[] for _ in range(unique_count)]

    for i, blade_index in enumerate(a_blade_indices):
        blade_index_set = frozenset(blade_index)
        if blade_index_set in blade_to_index:
            index = blade_to_index[blade_index_set]
            sign, _ = reduce_bases(blade_index, blade_to_blade_index[blade_index_set], positive_signature)
        else:
            index = len(blade_to_index)
            sign = 1
            blade_to_index[blade_index_set] = index
            blade_to_blade_index[blade_index_set] = blade_index
            indices.append(blade_index)
        out_indices[index].append(i)
        out_signs[index].append(sign)

    def _values_mv_reduce_same(a_values):
        out_batch_shape = a_values.shape[1:]
        result = jnp.empty([len(out_indices), *out_batch_shape], dtype=jnp.float32)
        for i, (mm, signs) in enumerate(zip(out_indices, out_signs)):
            for j, (m, sign) in enumerate(zip(mm, signs)):
                if j == 0:
                    result = result.at[i].set(sign * a_values[m])
                else:
                    result = result.at[i].add(sign * a_values[m])
        return result
    _values_mv_reduce_same_jit = jax.jit(_values_mv_reduce_same)

    return _values_mv_reduce_same_jit, tuple(indices)



import jax
from functools import cache
from .multiply import get_mv_multiply
from .select import get_mv_select
from .add import get_mv_add
from ..jaxga import is_scalar_index


@cache
def get_mv_inverse(a_blade_indices, signature):
    dims = len(set(a_blade_indices))
    n = 2 ** ((dims + 1) // 2)

    last_ind = a_blade_indices

    selects = []
    adds = []
    mults = []
    for k in range(1, n):
        select, select_ind = get_mv_select(last_ind, is_scalar_index)
        add, add_ind = get_mv_add(last_ind, select_ind)
        mult, mult_ind = get_mv_multiply(a_blade_indices, add_ind, signature)

        selects.append(select)
        adds.append(add)
        mults.append(mult)

        last_ind = mult_ind

    select_last_scalar, _ = get_mv_select(last_ind, is_scalar_index)

    def _values_mv_inverse(a_values):
        u = a_values
        for k, (select, add, mult) in enumerate(zip(selects, adds, mults)):
            c = n / (k + 1) * select(u)
            u_minus_c = add(u, -c)
            u = mult(a_values, u_minus_c)

        return u_minus_c / select_last_scalar(u)
    _values_mv_inverse_jit = jax.jit(_values_mv_inverse)

    return _values_mv_inverse_jit, add_ind



import itertools
import jax
import jax.numpy as jnp
from ..jaxga import reduce_bases
from functools import cache


@cache
def get_mv_multiply(a_blade_indices, b_blade_indices, signature, prod="gp"):
    out_indices = []
    out_blade_indices = []
    out_signs = []
    out_indices = []
    indices_a = []
    indices_b = []

    blade_to_index = {}

    for (i_a, index_a), (i_b, index_b) in itertools.product(
        enumerate(a_blade_indices),
        enumerate(b_blade_indices)
    ):
        out_sign, out_index = reduce_bases(index_a, index_b, signature)
        if out_sign != 0 and (
            prod == "gp" or
            (prod == "op" and len(out_index) == abs(len(index_a) + len(index_b))) or
            (prod == "ip" and len(out_index) == abs(len(index_a) - len(index_b)))
        ):
            out_signs.append(out_sign)
            indices_a.append(i_a)
            indices_b.append(i_b)

            if out_index in blade_to_index:
                out_indices.append(blade_to_index[out_index])
            else:
                blade_to_index[out_index] = len(blade_to_index)
                out_indices.append(blade_to_index[out_index])
                out_blade_indices.append(out_index)

    if len(out_indices) == 0:
        def _values_mv_mul(a_values, b_values):
            return jnp.zeros((), dtype=jnp.float32)
    else:
        out_size = max(out_indices) + 1

        def _values_mv_mul(a_values, b_values):
            out_batch_shape = jnp.broadcast_shapes(
                a_values.shape[1:], b_values.shape[1:]
            )
            out_values = jnp.zeros(
                [out_size, *out_batch_shape], dtype=jnp.float32
            )

            for index_a, index_b, out_sign, out_index in zip(indices_a, indices_b, out_signs, out_indices):
                out_values = out_values.at[out_index].add(
                    out_sign * a_values[index_a] * b_values[index_b]
                )

            return out_values

    _values_mv_mul_jit = jax.jit(_values_mv_mul, static_argnames=["out_size"])

    return _values_mv_mul_jit, tuple(out_blade_indices)



from functools import cache
from .signatures import positive_signature


def blade_name(blade_indices):
    return "e_{%s}" % ", ".join(str(ind) for ind in blade_indices)


def mv_repr(indices, values):
    return "Multivector(%s)" % " + ".join(
        "%s %s" % (val, blade_name(ind))
        for (ind, val) in zip(indices, values)
    )


def _normal_swap(x):
    for i in range(len(x) - 1):
        a, b = x[i], x[i + 1]
        if a > b:
            x[i], x[i+1] = b, a
            return False, x
    return True, x


def get_normal_ordered(blade_name):
    blade_name = list(blade_name)
    sign = -1
    done = False
    while not done:
        sign *= -1
        done, blade_name = _normal_swap(blade_name)
    return sign, tuple(blade_name)


def _collapse_same(x):
    for i in range(len(x) - 1):
        a, b = x[i], x[i + 1]
        if a == b:
            return False, x[:i] + x[i+2:], a
    return True, x, None


@cache
def reduce_bases(a, b, metric):
    combined = list(a + b)

    # Bring into normal order:
    sign, combined = get_normal_ordered(combined)

    done = False
    while not done:
        done, combined, combined_elem = _collapse_same(combined)
        if not done:
            sign *= metric(combined_elem)

    return sign, tuple(combined)


def is_scalar_index(b):
    return b == ()


def pseudoscalar_index(dims):
    return tuple(range(dims))


def dual_blade_index(blade_index, dims):
    pss = pseudoscalar_index(dims)
    blade_index_reverse = tuple(reversed(blade_index))
    return reduce_bases(blade_index_reverse, pss, positive_signature)


def reverse_indices(blade_indices):
    return [tuple(reversed(blade_index)) for blade_index in blade_indices]



import jax
import jax.numpy as jnp
from functools import cache


@cache
def get_mv_select(a_blade_indices, select_index):
    out_blade_indices = []
    out_a_indices = []
    for i, a_index in enumerate(a_blade_indices):
        if select_index(a_index):
            out_blade_indices.append(a_index)
            out_a_indices.append(i)
    out_a_indices = jnp.array(out_a_indices, dtype=jnp.int32)

    def _values_mv_select(a_values):
        return a_values[out_a_indices]

    _values_mv_select_jit = jax.jit(_values_mv_select)

    return _values_mv_select_jit, tuple(out_blade_indices)



from .ops.inverse import get_mv_inverse
from .ops.dual import get_mv_dual
from .ops.reduce_same import get_mv_reduce_same
from .ops.keepnonzero import get_mv_keep_nonzero
from .ops.multiply import get_mv_multiply
from .ops.add import get_mv_add
from .ops.simple_exp import get_mv_simple_exp
from .ops.select import get_mv_select
from .jaxga import reverse_indices, mv_repr
from .signatures import positive_signature
import jax.numpy as jnp


class MultiVector:
    def e(*indices, **kwargs):
        return MultiVector(
            values=jnp.ones([1], dtype=jnp.float32),
            indices=(tuple(indices),),
            signature=kwargs["signature"] if "signature" in kwargs else positive_signature
        )

    def __init__(self, values, indices, signature=positive_signature):
        self.values = values
        self.indices = tuple(indices)
        self.signature = signature

    def __add__(self, other):
        mv_add, out_indices = get_mv_add(self.indices, other.indices)
        out_values = mv_add(self.values, other.values)
        return MultiVector(values=out_values, indices=out_indices, signature=self.signature)

    def __sub__(self, other):
        mv_add, out_indices = get_mv_add(self.indices, other.indices)
        out_values = mv_add(self.values, -other.values)
        return MultiVector(values=out_values, indices=out_indices, signature=self.signature)

    def __mul__(self, other):
        if isinstance(other, MultiVector):
            mv_multiply, out_indices = get_mv_multiply(
                self.indices, other.indices, self.signature)
            out_values = mv_multiply(self.values, other.values)
            return MultiVector(values=out_values, indices=out_indices, signature=self.signature)
        return MultiVector(values=self.values * other, indices=self.indices, signature=self.signature)

    def __rmul__(self, other):
        if isinstance(other, MultiVector):
            mv_multiply, out_indices = get_mv_multiply(
                other.indices, self.indices, self.signature)
            out_values = mv_multiply(other.values, self.values)
            return MultiVector(values=out_values, indices=out_indices)
        return MultiVector(values=self.values * other, indices=self.indices, signature=self.signature)

    def __xor__(self, other):
        mv_multiply, out_indices = get_mv_multiply(
            self.indices, other.indices, self.signature, "op")
        out_values = mv_multiply(self.values, other.values)
        return MultiVector(values=out_values, indices=out_indices, signature=self.signature)

    def __or__(self, other):
        mv_multiply, out_indices = get_mv_multiply(
            self.indices, other.indices, self.signature, "ip")
        out_values = mv_multiply(self.values, other.values)
        return MultiVector(values=out_values, indices=out_indices, signature=self.signature)

    def __invert__(self):
        return MultiVector(values=self.values, indices=reverse_indices(self.indices), signature=self.signature)

    def __neg__(self):
        return MultiVector(values=-self.values, indices=reverse_indices(self.indices), signature=self.signature)

    def inverse(self):
        mv_inv, inv_indices = get_mv_inverse(self.indices, self.signature)
        inv_values = mv_inv(self.values)
        return MultiVector(values=inv_values, indices=inv_indices, signature=self.signature)

    def __truediv__(self, other):
        if isinstance(other, MultiVector):
            return self * other.inverse()
        return self * (1 / other)

    def __rtruediv__(self, other):
        return other * self.inverse()

    def __repr__(self):
        return mv_repr(self.indices, self.values)

    def __getitem__(self, select_indices):
        mv_select, out_indices = get_mv_select(self.indices, select_indices)
        out_values = mv_select(self.values)
        return MultiVector(out_values, out_indices, signature=self.signature)

    def simple_exp(self):
        mv_simple_exp, out_indices = get_mv_simple_exp(
            self.indices, self.signature)
        out_values = mv_simple_exp(self.values)
        return MultiVector(out_values, out_indices, signature=self.signature)

    def keep_nonzero(self):
        mv_keep_nonzero, out_indices = get_mv_keep_nonzero(
            self.indices, self.values
        )
        out_values = mv_keep_nonzero(self.values)
        return MultiVector(out_values, out_indices, signature=self.signature)

    def reduce_same(self):
        mv_reduce_same, out_indices = get_mv_reduce_same(
            self.indices
        )
        out_values = mv_reduce_same(self.values)
        return MultiVector(out_values, out_indices, signature=self.signature)

    def dual(self, dims):
        mv_dual, out_indices = get_mv_dual(self.indices, dims)
        out_values = mv_dual(self.values)
        return MultiVector(out_values, out_indices, signature=self.signature)
